#include <erl_driver.h>
#include <sys/types.h>
#include <grp.h>
#include <pwd.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>

typedef struct {
	ErlDrvPort port;
} erl_groupcheck_data;

static ErlDrvData erl_groupcheck_drv_start(ErlDrvPort port, char *buf) {
	erl_groupcheck_data *data = (erl_groupcheck_data *)driver_alloc(sizeof(erl_groupcheck_data));

	data->port = port;

	return (ErlDrvData)data;
}

static void erl_groupcheck_drv_stop(ErlDrvData handle) {
	driver_free((char *)handle);
}

static void copy_string_n(char **dest, char *src, size_t n) {
	*dest = malloc(sizeof(char) * (n + 1));
	strncpy(*dest, src, n);
	(*dest)[n] = '\0';
}

static void erl_groupcheck_drv_output(ErlDrvData handle, char *buf, int buflen) {
	erl_groupcheck_data *data = (erl_groupcheck_data *)handle;
#ifdef WIN32
	/* TODO Actually write Windows groupcheck equivalent code? */
	char res = 1;
#else
	char **member;
	int grouplen, userlen;
	struct group *group = NULL;
	struct passwd *passwd = NULL;
	char *groupname, *username, *sep, res = 0;

	sep = strchr(buf, ':');
	grouplen = sep-buf;
	copy_string_n(&groupname, buf, grouplen);
	sep++;

	userlen = buflen-(grouplen+1);
	copy_string_n(&username, sep, userlen);

	passwd = getpwnam(username);
	if (passwd == NULL) {
		driver_output(data->port, &res, 1);
		free(username);
		free(groupname);
		return;
	}

	group = getgrnam(groupname);
	if (group == NULL) {
		driver_output(data->port, &res, 1);
		free(username);
		free(groupname);
		return;
	}

	/* This is for groups without explicit member lists.
	 * Users in such a group have their primary gid set to the group gid. */
	res = (passwd->pw_gid == group->gr_gid);

	if (res == 0) {
		for (member = group->gr_mem; *member; ++member) {
			if (strcmp(username, *member) == 0) {
				res = 1;
				break;
			}
		}
	}

	free(username);
	free(groupname);
#endif

	driver_output(data->port, &res, 1);
}

ErlDrvEntry erl_groupcheck_entry = {
	NULL,                      /* F_PTR init, N/A */
	erl_groupcheck_drv_start,  /* L_PTR start, called when port is opened */
	erl_groupcheck_drv_stop,   /* F_PTR stop, called when port is closed */
	erl_groupcheck_drv_output, /* F_PTR output, called when erlang has sent */
	NULL,                      /* F_PTR ready_input, called when input description ready */
	NULL,                      /* F_PTR ready_output, called when output descriptor ready */
	"erl_groupcheck",          /* char *driver_name, the argument to open_port */
	NULL,                      /* F_PTR finish, called when unloaded */
	NULL,                      /* F_PTR control, port_command callback */
	NULL,                      /* F_PTR timeout, reserved */
	NULL                       /* F_PTR outputv, reserved */
};

DRIVER_INIT(erl_groupcheck) {
	return &erl_groupcheck_entry;
}
